from copy import deepcopy
from unittest.mock import Mock, MagicMock, call
from pylxd.exceptions import NotFound
from yaml import safe_dump

from vnet_manager.tests import VNetTestCase
from vnet_manager.conf import settings
from vnet_manager.operations.machine import (
    show_status,
    check_if_lxc_machine_exists,
    get_lxc_machine_status,
    wait_for_lxc_machine_status,
    change_machine_status,
    change_lxc_machine_status,
    create_machines,
    create_lxc_machines_from_base_image,
    destroy_machines,
    destroy_lxc_machine,
    place_lxc_interface_configuration_on_container,
    generate_machine_netplan_config,
    create_lxc_base_image_container,
    enable_type_specific_machine_configuration,
    configure_lxc_ip_forwarding,
)


class TestShowStatus(VNetTestCase):
    def setUp(self) -> None:
        self.tabulate = self.set_up_patch("vnet_manager.operations.machine.tabulate")
        self.get_lxc_machine_status = self.set_up_patch("vnet_manager.operations.machine.get_lxc_machine_status")
        self.get_lxc_machine_status.return_value = ["router", "up", "LXC"]
        self.config = deepcopy(settings.CONFIG)
        # Only 1 machine for less output
        self.config["machines"].pop("router101", None)
        self.config["machines"].pop("host102", None)

    def test_show_status_call_get_lxc_machine_status(self):
        show_status(self.config)
        self.get_lxc_machine_status.assert_called_once_with("router100")

    def test_show_status_makes_correct_tabulate_call(self):
        show_status(self.config)
        self.tabulate.assert_called_once_with(
            [self.get_lxc_machine_status.return_value], headers=["Name", "Status", "Provider"], tablefmt="pretty"
        )


class TestCheckIfLXCMachineExists(VNetTestCase):
    def setUp(self) -> None:
        self.machine = Mock()
        self.lxd_client = self.set_up_patch("vnet_manager.operations.machine.get_lxd_client")
        self.lxd_client.return_value = self.machine

    def test_check_if_lxc_machine_exists_calls_exists_method(self):
        check_if_lxc_machine_exists("test")
        self.machine.containers.exists.assert_called_once_with("test")

    def test_check_if_lxc_machine_exists_returns_value_of_exists_method(self):
        self.machine.containers.exists.return_value = False
        self.assertFalse(check_if_lxc_machine_exists("test"))


class TestGetLXCMachineStatus(VNetTestCase):
    def setUp(self) -> None:
        self.machine = Mock()
        self.lxd_client = self.set_up_patch("vnet_manager.operations.machine.get_lxd_client")
        self.lxd_client.return_value = self.machine

    def test_get_lxc_machine_calls_lxd_client(self):
        get_lxc_machine_status("test")
        self.lxd_client.assert_called_once_with()

    def test_get_lxc_machine_status_calls_get_method(self):
        get_lxc_machine_status("test")
        self.machine.containers.get.assert_called_once_with("test")

    def test_get_lxc_machine_status_returns_list_with_status_of_container(self):
        self.machine.containers.get.return_value.status = "banaan"
        self.assertEqual(get_lxc_machine_status("test"), ["test", "banaan", "LXC"])

    def test_get_lxc_machine_status_returns_na_on_not_found_exception(self):
        self.machine.containers.get.side_effect = NotFound(response="blaap")
        self.assertEqual(get_lxc_machine_status("test"), ["test", "NA", "LXC"])


class TestWaitForLXCMachineStatus(VNetTestCase):
    def setUp(self) -> None:
        self.machine = MagicMock()
        self.machine.state.return_value.status = "123"
        self.sleep = self.set_up_patch("vnet_manager.operations.machine.sleep")

    def test_wait_for_lxc_machine_status_calls_state(self):
        wait_for_lxc_machine_status(self.machine, "123")
        self.machine.state.assert_called_once_with()

    def test_wait_for_lxc_machine_status_does_not_call_sleep_if_state_correct(self):
        wait_for_lxc_machine_status(self.machine, "123")
        self.assertFalse(self.sleep.called)

    def test_wait_for_lxc_machine_status_raises_timeout_error_if_right_status_was_not_returned(self):
        self.machine.state.return_value.status = "456"
        with self.assertRaises(TimeoutError):
            wait_for_lxc_machine_status(self.machine, "123")
        self.assertTrue(self.sleep.called)

    def test_wait_for_lxc_machine_status_deals_with_weird_capitals(self):
        self.machine.state.return_value.status = "RuNNING"
        wait_for_lxc_machine_status(self.machine, "rUNNing")


class TestChangeMachineStatus(VNetTestCase):
    def setUp(self) -> None:
        self.change_lxc_machine_status = self.set_up_patch("vnet_manager.operations.machine.change_lxc_machine_status")

    def test_change_machine_status_raises_not_implemented_error_when_invalid_status(self):
        with self.assertRaises(NotImplementedError):
            change_machine_status(settings.CONFIG, status="blaap")

    def test_change_machine_status_calls_change_lxc_machine_status_with_machines_from_config(self):
        change_machine_status(settings.CONFIG)
        calls = [call(m, status="stop") for m in settings.CONFIG["machines"].keys()]
        self.change_lxc_machine_status.assert_has_calls(calls)

    def test_change_machine_status_calls_change_lxc_machine_status_only_with_given_machine_list(self):
        machines = ["router100"]
        change_machine_status(settings.CONFIG, machines=machines)
        self.change_lxc_machine_status.assert_called_once_with(machines[0], status="stop")

    def test_change_machine_status_calls_change_lxc_machine_status_with_different_status(self):
        change_machine_status(settings.CONFIG, status="start")
        calls = [call(m, status="start") for m in settings.CONFIG["machines"].keys()]
        self.change_lxc_machine_status.assert_has_calls(calls)

    def test_change_machine_status_skips_non_existent_machines(self):
        machines = ["banaan1"]
        change_machine_status(settings.CONFIG, machines=machines)
        self.assertFalse(self.change_lxc_machine_status.called)


class TestChangeLXCMachineStatus(VNetTestCase):
    def setUp(self) -> None:
        self.lxd_client = self.set_up_patch("vnet_manager.operations.machine.get_lxd_client")
        self.client = Mock()
        self.machine = Mock()
        self.lxd_client.return_value = self.client
        self.client.containers.get.return_value = self.machine
        self.sleep = self.set_up_patch("vnet_manager.operations.machine.sleep")
        self.wait_for_lxc_machine_status = self.set_up_patch("vnet_manager.operations.machine.wait_for_lxc_machine_status")

    def test_change_lxc_machine_status_calls_lxd_client(self):
        change_lxc_machine_status("banaan")
        self.lxd_client.assert_called_once_with()

    def test_change_lxc_machine_status_calls_sleep(self):
        change_lxc_machine_status("banaan")
        self.sleep.assert_called_once_with(1)

    def test_change_lxc_machine_status_calls_containers_get(self):
        change_lxc_machine_status("banaan")
        self.client.containers.get.assert_called_once_with("banaan")

    def test_change_lxc_machine_status_calls_machine_stop_by_default(self):
        change_lxc_machine_status("banaan")
        self.machine.stop.assert_called_once_with()

    def test_change_lxc_machine_status_does_not_call_start_by_default(self):
        change_lxc_machine_status("banaan")
        self.assertFalse(self.machine.start.called)

    def test_change_lxc_machine_status_deals_with_not_found_error(self):
        self.client.containers.get.side_effect = NotFound(response="blaap")
        change_lxc_machine_status("banaan")
        self.assertFalse(self.machine.stop.called)
        self.assertFalse(self.machine.start.called)
        self.assertFalse(self.wait_for_lxc_machine_status.called)

    def test_change_lxc_machine_status_call_start_when_passed(self):
        change_lxc_machine_status("banaan", status="start")
        self.machine.start.assert_called_once_with(wait=True)

    def test_change_lxc_machine_status_does_not_call_stop_when_start_passed(self):
        change_lxc_machine_status("banaan", status="start")
        self.assertFalse(self.machine.stop.called)

    def test_change_lxc_machine_status_calls_wait_for_lxc_machine_status_with_stopped(self):
        change_lxc_machine_status("banaan", status="stop")
        self.wait_for_lxc_machine_status.assert_called_once_with(self.machine, "Stopped")

    def test_change_lxc_machine_status_calls_wait_for_lxc_machine_status_with_running(self):
        change_lxc_machine_status("banaan", status="start")
        self.wait_for_lxc_machine_status.assert_called_once_with(self.machine, "Running")

    def test_change_lxc_machine_status_catches_timeout_error(self):
        self.wait_for_lxc_machine_status.side_effect = TimeoutError()
        change_lxc_machine_status("banaan")


class TestCreateMachines(VNetTestCase):
    def setUp(self) -> None:
        self.create_lxc_machines_from_base_image = self.set_up_patch("vnet_manager.operations.machine.create_lxc_machines_from_base_image")

    def test_create_machines_calls_create_lxc_machines_with_config_machines(self):
        create_machines(settings.CONFIG)
        self.create_lxc_machines_from_base_image.assert_called_once_with(settings.CONFIG, settings.CONFIG["machines"].keys())

    def test_create_machines_calls_create_lxc_machine_with_custom_machine_list(self):
        create_machines(settings.CONFIG, machines=["test1", "test2"])
        self.create_lxc_machines_from_base_image.assert_called_once_with(settings.CONFIG, ["test1", "test2"])


class TestCreateLXCMachinesFromBaseImage(VNetTestCase):
    def setUp(self) -> None:
        self.check_if_lxc_machine_exists = self.set_up_patch("vnet_manager.operations.machine.check_if_lxc_machine_exists")
        self.check_if_lxc_machine_exists.return_value = False
        self.lxd_client = self.set_up_patch("vnet_manager.operations.machine.get_lxd_client")
        self.client = Mock()
        self.lxd_client.return_value = self.client
        self.place_lxc_interface_configuration_on_container = self.set_up_patch(
            "vnet_manager.operations.machine.place_lxc_interface_configuration_on_container"
        )
        self.request_confirm = self.set_up_patch("vnet_manager.operations.machine.request_confirmation")
        self.config = deepcopy(settings.CONFIG)
        self.excepted_config = {
            "name": "router100",
            "source": {"alias": settings.LXC_BASE_IMAGE_ALIAS, "type": "image"},
            "ephemeral": False,
            "config": {"user.network-config": "disabled"},
            "devices": {
                "eth0": {"type": "none"},
                "eth12": {
                    "name": "eth12",
                    "host_name": "router100-eth12",
                    "parent": "vnet-br0",
                    "type": "nic",
                    "nictype": "bridged",
                    "hwaddr": "00:00:00:00:01:11",
                },
            },
            "profiles": [settings.LXC_VNET_PROFILE],
        }

    def test_create_lxc_machines_from_base_image_calls_lxc_client(self):
        create_lxc_machines_from_base_image(settings.CONFIG, ["router100", "router101"])
        self.lxd_client.assert_called_once_with()

    def test_create_lxc_machines_from_base_image_does_nothing_if_machines_already_exist(self):
        self.check_if_lxc_machine_exists.return_value = True
        create_lxc_machines_from_base_image(self.config, ["router100", "router101"])
        self.assertFalse(self.place_lxc_interface_configuration_on_container.called)
        self.assertFalse(self.client.containers.create.called)

    def test_create_lxc_machines_from_base_image_calls_containers_create_method(self):
        create_lxc_machines_from_base_image(self.config, ["router100"])
        self.client.containers.create.assert_called_once_with(self.excepted_config, wait=True)

    def test_create_lxc_machines_from_base_image_calls_place_lxc_interface_configuration_on_container(self):
        create_lxc_machines_from_base_image(self.config, ["router100"])
        self.place_lxc_interface_configuration_on_container.assert_called_once_with(self.config, "router100")

    def test_create_lxc_machine_from_base_image_calls_create_functions_on_the_basis_of_the_number_of_containers(self):
        create_lxc_machines_from_base_image(self.config, ["router100", "router101", "host102"])
        self.assertEqual(self.client.containers.create.call_count, 3)
        self.assertEqual(self.place_lxc_interface_configuration_on_container.call_count, 3)

    def test_create_lxc_machines_from_base_image_calls_request_confirmation_if_containers_already_created(self):
        self.check_if_lxc_machine_exists.return_value = True
        create_lxc_machines_from_base_image(self.config, ["router100", "host102"])
        self.request_confirm.assert_called_once_with(
            message="Some containers already existed, the next operation will overwrite network, "
            "host and user config files on those containers"
        )


class TestDestroyMachines(VNetTestCase):
    def setUp(self) -> None:
        self.request_confirm = self.set_up_patch("vnet_manager.operations.machine.request_confirmation")
        self.destroy_lxc_machine = self.set_up_patch("vnet_manager.operations.machine.destroy_lxc_machine")

    def test_destroy_machines_calls_request_confirmation(self):
        destroy_machines(settings.CONFIG)
        self.request_confirm.assert_called_once_with(
            message="Requesting confirmation of deletion for the following machines: router100, router101, host102",
            prompt="This operation cannot be undone. Are you sure?! (yes/no) ",
        )

    def test_destroy_machines_calls_destroy_lxc_machine_per_lxc_machine(self):
        calls = [call(m) for m in settings.CONFIG["machines"].keys()]
        destroy_machines(settings.CONFIG)
        self.destroy_lxc_machine.assert_has_calls(calls)

    def test_destroy_machines_calls_destroy_lxc_machine_with_specific_machine(self):
        destroy_machines(settings.CONFIG, machines=["router100"])
        self.destroy_lxc_machine.assert_called_once_with("router100")


class TestDestroyLXCMachine(VNetTestCase):
    def setUp(self) -> None:
        self.lxd_client = self.set_up_patch("vnet_manager.operations.machine.get_lxd_client")
        self.client = Mock()
        self.machine = Mock()
        self.lxd_client.return_value = self.client
        self.client.containers.get.return_value = self.machine

    def test_destroy_lxc_machine_calls_lxd_client(self):
        destroy_lxc_machine("banaan")
        self.lxd_client.assert_called_once_with()

    def test_destroy_lxc_machine_calls_container_delete_method(self):
        destroy_lxc_machine("banaan")
        self.machine.delete.assert_called_once_with(wait=False)

    def test_destroy_lxc_machine_does_not_call_stop_by_default(self):
        destroy_lxc_machine("banaan")
        self.assertFalse(self.machine.stop.called)

    def test_destroy_lxc_machine_calls_machine_stop_when_status_is_running(self):
        self.machine.status = "Running"
        destroy_lxc_machine("banaan")
        self.machine.stop.assert_called_once_with(wait=True)

    def test_destroy_lxc_machine_does_nothing_if_machine_does_not_exist(self):
        self.client.containers.get.side_effect = NotFound(response="blaap")
        destroy_lxc_machine("banaan")
        self.assertFalse(self.machine.delete.called)
        self.assertFalse(self.machine.stop.called)


class TestCreateLXCBaseImageContainer(VNetTestCase):
    def setUp(self) -> None:
        self.check_if_machine_exists = self.set_up_patch("vnet_manager.operations.machine.check_if_lxc_machine_exists")
        self.check_if_machine_exists.return_value = False
        self.request_confirm = self.set_up_patch("vnet_manager.operations.machine.request_confirmation")
        self.destroy_lxc_machine = self.set_up_patch("vnet_manager.operations.machine.destroy_lxc_machine")
        self.lxd_client = self.set_up_patch("vnet_manager.operations.machine.get_lxd_client")
        self.client = Mock()
        self.lxd_client.return_value = self.client

    def test_create_lxc_base_image_container_calls_check_if_machine_exists(self):
        create_lxc_base_image_container()
        self.check_if_machine_exists.assert_called_once_with(settings.LXC_BASE_IMAGE_MACHINE_NAME)

    def test_create_lxc_base_image_container_does_not_call_destroy_lxc_machine_when_machine_does_not_exist(self):
        create_lxc_base_image_container()
        self.assertFalse(self.request_confirm.called)
        self.assertFalse(self.destroy_lxc_machine.called)

    def test_create_lxc_base_image_container_calls_confirm_when_machine_already_exists(self):
        self.check_if_machine_exists.return_value = True
        create_lxc_base_image_container()
        self.request_confirm.assert_called_once_with(
            message="Recreating it will destroy any local changes", prompt="Recreate the LXC base image machine? "
        )

    def test_create_lxc_base_image_container_calls_destroy_lxc_machine_when_machine_already_exists(self):
        self.check_if_machine_exists.return_value = True
        create_lxc_base_image_container()
        self.destroy_lxc_machine.assert_called_once_with(settings.LXC_BASE_IMAGE_MACHINE_NAME, wait=True)

    def test_create_lxc_base_image_calls_get_lxd_client(self):
        create_lxc_base_image_container()
        self.lxd_client.assert_called_once_with()

    def test_create_lxc_base_image_calls_client_create_function(self):
        excepted_config = {
            "name": settings.LXC_BASE_IMAGE_MACHINE_NAME,
            "architecture": "x86_64",
            "profiles": [settings.LXC_VNET_PROFILE],
            "ephemeral": False,
            "config": {},
            "devices": {
                "eth0": {
                    "name": "eth0",
                    "parent": "lxdbr0",
                    "type": "nic",
                    "nictype": "bridged",
                    "host_name": "{}-eth0".format(settings.LXC_BASE_IMAGE_MACHINE_NAME),
                }
            },
            "source": {
                "type": "image",
                "protocol": str(settings["PROVIDERS"]["lxc"]["base_image"]["protocol"]),
                "server": settings["PROVIDERS"]["lxc"]["base_image"]["server"],
                "alias": str(settings["PROVIDERS"]["lxc"]["base_image"]["os"]),
            },
        }
        create_lxc_base_image_container()
        self.client.containers.create.assert_called_once_with(excepted_config, wait=True)


class TestPlaceLXCInterfaceConfigurationOnContainer(VNetTestCase):
    def setUp(self) -> None:
        self.network_conf = self.set_up_patch("vnet_manager.operations.machine.generate_machine_netplan_config")
        self.network_conf.return_value = {"int": "456"}
        self.write_files_to_lxc_container = self.set_up_patch("vnet_manager.operations.machine.write_file_to_lxc_container")
        self.config = {"int": "123"}

    def test_place_lxc_interface_configuration_on_container_calls_generate_machine_netplan_config(self):
        place_lxc_interface_configuration_on_container(self.config, "router100")
        self.network_conf.assert_called_once_with(self.config, "router100")

    def test_place_lxc_interface_configuration_on_container_calls_write_file_to_lxc_container(self):
        place_lxc_interface_configuration_on_container(self.config, "router100")
        self.write_files_to_lxc_container.assert_called_once_with(
            "router100", settings.VNET_NETPLAN_CONFIG_FILE_PATH, safe_dump(self.network_conf.return_value)
        )


class TestEnableTypeSpecificMachineConfiguration(VNetTestCase):
    def setUp(self) -> None:
        self.configure_lxc_ip_forwarding = self.set_up_patch("vnet_manager.operations.machine.configure_lxc_ip_forwarding")

    def test_enable_type_specific_machine_configuration_calls_configure_lxc_forwarding_for_each_machine(self):
        enable_type_specific_machine_configuration(settings.CONFIG)
        self.configure_lxc_ip_forwarding.assert_has_calls(
            [call("router100", enable=True), call("router101", enable=True), call("host102", enable=False)]
        )
        self.assertEqual(self.configure_lxc_ip_forwarding.call_count, 3)


class TestConfigureLXCIPForwarding(VNetTestCase):
    def setUp(self) -> None:
        self.write_file_to_lxc_container = self.set_up_patch("vnet_manager.operations.machine.write_file_to_lxc_container")

    def test_configure_lxc_ip_forwarding_calls_write_file_to_lxc_container(self):
        configure_lxc_ip_forwarding("router100")
        self.write_file_to_lxc_container.assert_has_calls(
            [
                call("router100", "/etc/sysctl.d/20-net.ipv4.ip_forward.conf", "net.ipv4.ip_forward=1\n"),
                call("router100", "/etc/sysctl.d/20-net.ipv6.conf.all.forwarding.conf", "net.ipv6.conf.all.forwarding=1\n"),
            ]
        )
        self.assertEqual(self.write_file_to_lxc_container.call_count, 2)


class TestGenerateMachineNetplanConfig(VNetTestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        self.config = deepcopy(settings.CONFIG)
        self.expected_config = {
            "network": {
                "version": 2,
                "renderer": "networkd",
                "ethernets": {
                    "eth12": {
                        "dhcp4": "no",
                        "dhcp6": "no",
                        "match": {"macaddress": "00:00:00:00:01:11"},
                        "set-name": "eth12",
                        "addresses": ["192.168.0.2/24", "fd00:12::2/64"],
                    },
                },
                "vlans": {"vlan.100": {"id": 100, "link": "eth12", "addresses": ["10.0.100.1/24"], "dhcp4": "no", "dhcp6": "no",}},
            }
        }

    def test_generate_machine_netplan_config_returns_a_dict(self):
        self.assertIsInstance(generate_machine_netplan_config(self.config, "router100"), dict)

    def test_generate_machine_netplan_config_returns_expected_config_without_vlans(self):
        del self.config["machines"]["router100"]["vlans"]
        del self.expected_config["network"]["vlans"]
        self.assertEqual(self.expected_config, generate_machine_netplan_config(self.config, "router100"))

    def test_generate_machine_netplan_config_returns_expected_config_with_vlans(self):
        self.assertEqual(self.expected_config, generate_machine_netplan_config(self.config, "router100"))



import logging

from vnet_manager.log import setup_console_logging
from vnet_manager.tests import VNetTestCase
from vnet_manager.conf import settings


class TestLog(VNetTestCase):
    def test_setup_console_logging_sets_up_INFO_logging_by_default(self):
        setup_console_logging()
        self.assertEqual(settings.LOGGING["handlers"]["console"]["level"], logging.INFO)

    def test_setup_console_logging_sets_up_passed_logging_level(self):
        setup_console_logging(logging.DEBUG)
        self.assertEqual(settings.LOGGING["handlers"]["console"]["level"], logging.DEBUG)



from unittest.mock import call

from vnet_manager.tests import VNetTestCase
from vnet_manager.conf import perform_import


class TestPerformImport(VNetTestCase):
    def setUp(self) -> None:
        self.mock_import_from_string = self.set_up_patch("vnet_manager.conf.import_from_string", return_value="imported")

    def test_perform_import_return_none_if_value_is_none(self):
        returned = perform_import(None, "abc")
        self.assertIsNone(returned)

    def test_perform_import_return_value_if_incorrect_val_passed(self):
        returned = perform_import(123, "xyz")
        self.assertEqual(returned, 123)

    def test_perform_import_return_import_result_if_single_import_is_performed(self):
        returned = perform_import("abc", "xyz")
        self.assertEqual(returned, "imported")

    def test_perform_import_return_import_result_list_if_import_list_is_performed(self):
        returned = perform_import(["a", "b", "c"], "xyz")
        self.assertListEqual(returned, ["imported", "imported", "imported"])

    def test_perform_import_that_import_from_string_is_called_once_on_string_value(self):
        perform_import("abc", "xyz")
        expected_calls = [call("abc", "xyz")]
        self.assertEqual(expected_calls, self.mock_import_from_string.mock_calls)

    def test_perform_import_that_import_from_list_is_called_for_each_element(self):
        perform_import(["a", "b", "c"], "xyz")
        expected_calls = [call("a", "xyz"), call("b", "xyz"), call("c", "xyz")]
        self.assertListEqual(expected_calls, self.mock_import_from_string.mock_calls)



import sys
from argparse import ArgumentParser
from logging import INFO, DEBUG, getLogger
from os import EX_NOPERM, environ

from vnet_manager.conf import settings
from vnet_manager.log import setup_console_logging
from vnet_manager.actions.manager import ActionManager
from vnet_manager.utils.user import check_for_root_user

logger = getLogger(__name__)


def parse_args(args=None):
    parser = ArgumentParser(description="VNet-manager a virtual network manager - manages containers to create virtual networks")
    parser.add_argument(
        "action",
        choices=sorted(settings.VALID_ACTIONS),
        help="The action to preform on the virtual network, use '<action> help' for information about that action",
    )
    parser.add_argument("config", help="The yaml config file to use", nargs="?", default="default")

    # Options
    parser.add_argument(
        "-m",
        "--machines",
        nargs="*",
        help="Just apply the actions on the following machine names " "(default is all machines defined in the config file)",
    )
    parser.add_argument("-y", "--yes", action="store_true", help="Answer yes to all questions")
    parser.add_argument("-v", "--verbose", action="store_true", help="Print debug messages")
    parser.add_argument("-nh", "--no-hosts", action="store_true", help="Disable creation of /etc/hosts")

    start_group = parser.add_argument_group("Start options", "These options can be specified for the start action")
    start_group.add_argument("-s", "--sniffer", action="store_true", help="Start a TCPdump sniffer on the VNet interfaces")
    destroy_group = parser.add_argument_group("Destroy options", "These options can be specified for the destroy action")
    destroy_group.add_argument("-b", "--base-image", action="store_true", help="Destroy the base image instead of the machines")
    args = parser.parse_args(args=args)

    # User input sanity checks
    if args.action == "status":
        # For people who are used to status calls
        args.action = "show"
    if args.config == "default" and args.action in settings.CONFIG_REQUIRED_ACTIONS:
        parser.error("This action requires a config file to be passed")
    if args.sniffer and not args.action == "start":
        parser.error("The sniffer option only makes sense with the 'start' action")
    if args.base_image and not args.action == "destroy":
        parser.error("The base_image option only makes sense with the 'destroy' action")
    if args.no_hosts and not args.action == "create":
        parser.error("The no_hosts option only makes sense with the 'create' action")
    return args


def main(args=None):
    """
    Program entry point
    :param list args: The pre-cooked arguments to pass to the ArgParser
    :return int: exit_code
    """
    args = parse_args(args)
    # Set the VNET_FORCE variable, if --yes is given this will answer yes to all questions
    environ[settings.VNET_FORCE_ENV_VAR] = "true" if args.yes else "false"
    # Setup logging
    setup_console_logging(verbosity=DEBUG if args.verbose else INFO)
    # Most VNet operation require root. So, do a root check
    if not check_for_root_user():
        logger.critical("This program should only be run as root")
        return EX_NOPERM
    # Let the action manager handle the rest
    manager = ActionManager(config_path=args.config, sniffer=args.sniffer, base_image=args.base_image, no_hosts=args.no_hosts)
    if args.machines:
        manager.machines = args.machines
    return manager.execute(args.action)


if __name__ == "__main__":
    sys.exit(main())



from copy import deepcopy
import shlex
from unittest.mock import call, MagicMock

from vnet_manager.tests import VNetTestCase
from vnet_manager.environment.lxc import ensure_vnet_lxc_environment, cleanup_vnet_lxc_environment, configure_lxc_base_machine
from vnet_manager.conf import settings


class TestEnsureVNetLXCEnvironment(VNetTestCase):
    def setUp(self) -> None:
        self.config = deepcopy(settings.CONFIG)
        self.check_for_supported_os = self.set_up_patch("vnet_manager.environment.lxc.check_for_supported_os")
        self.check_for_installed_packages = self.set_up_patch("vnet_manager.environment.lxc.check_for_installed_packages")
        self.check_if_lxc_storage_pool_exists = self.set_up_patch("vnet_manager.environment.lxc.check_if_lxc_storage_pool_exists")
        self.create_lxc_storage_pool = self.set_up_patch("vnet_manager.environment.lxc.create_lxc_storage_pool")
        self.check_if_lxc_profile_exists = self.set_up_patch("vnet_manager.environment.lxc.check_if_lxc_profile_exists")
        self.create_vnet_lxc_profile = self.set_up_patch("vnet_manager.environment.lxc.create_vnet_lxc_profile")
        self.check_if_lxc_image_exists = self.set_up_patch("vnet_manager.environment.lxc.check_if_lxc_image_exists")
        self.create_lxc_base_image_container = self.set_up_patch("vnet_manager.environment.lxc.create_lxc_base_image_container")
        self.change_lxc_machine_status = self.set_up_patch("vnet_manager.environment.lxc.change_lxc_machine_status")
        self.configure_lxc_base_machine = self.set_up_patch("vnet_manager.environment.lxc.configure_lxc_base_machine")
        self.create_lxc_image_from_container = self.set_up_patch("vnet_manager.environment.lxc.create_lxc_image_from_container")
        self.destroy_lxc_machine = self.set_up_patch("vnet_manager.environment.lxc.destroy_lxc_machine")
        self.logger = self.set_up_patch("vnet_manager.environment.lxc.logger")

    def test_ensure_vnet_lxc_environment_does_nothing_if_no_lxc_machines_in_config(self):
        self.config["machines"] = {}
        ensure_vnet_lxc_environment(self.config)
        self.logger.debug.assert_called_once_with("Skipping LXC environment creation, no LXC machines in config")
        self.assertFalse(self.create_lxc_storage_pool.called)

    def test_ensure_vnet_lxc_environment_throws_runtime_error_when_running_on_unsupported_os(self):
        self.check_for_supported_os.return_value = False
        with self.assertRaises(RuntimeError):
            ensure_vnet_lxc_environment(self.config)
        self.logger.critical.assert_called_once_with("Unable to create LXC environment on your machine, OS not supported")

    def test_ensure_vnet_lxc_environment_throws_runtime_error_when_check_for_installed_packages_fails(self):
        self.check_for_installed_packages.return_value = False
        with self.assertRaises(RuntimeError):
            ensure_vnet_lxc_environment(self.config)
        self.logger.critical.assert_called_once_with(
            "Not all required host packages seem to be installed, please fix this before proceeding"
        )

    def test_ensure_vnet_lxc_environment_does_not_call_create_vnet_storage_pool_if_it_exists(self):
        self.check_if_lxc_storage_pool_exists.return_value = True
        ensure_vnet_lxc_environment(self.config)
        self.assertFalse(self.create_lxc_storage_pool.called)
        self.logger.debug.has_calls(call("VNet LXC storage pool {} found".format(settings.LXC_STORAGE_POOL_NAME)))

    def test_ensure_vnet_lxc_environment_calls_create_vnet_storage_pool_if_does_not_exist(self):
        self.check_if_lxc_storage_pool_exists.return_value = False
        ensure_vnet_lxc_environment(self.config)
        self.create_lxc_storage_pool.assert_called_once_with(name=settings.LXC_STORAGE_POOL_NAME, driver=settings.LXC_STORAGE_POOL_DRIVER)

    def test_ensure_vnet_lxc_environment_does_not_call_create_vnet_profile_if_it_exists(self):
        self.check_if_lxc_profile_exists.return_value = True
        ensure_vnet_lxc_environment(self.config)
        self.assertFalse(self.create_vnet_lxc_profile.called)
        self.logger.debug.has_calls(call("VNet profile {} found".format(settings.LXC_VNET_PROFILE)))

    def test_ensure_vnet_lxc_environment_calls_create_vnet_lxc_profile_if_it_does_not_exist(self):
        self.check_if_lxc_profile_exists.return_value = False
        ensure_vnet_lxc_environment(self.config)
        self.create_vnet_lxc_profile.assert_called_once_with(settings.LXC_VNET_PROFILE)

    def test_ensure_vnet_lxc_environment_does_not_call_base_image_creation_functions_when_it_exists(self):
        self.check_if_lxc_image_exists.return_value = True
        ensure_vnet_lxc_environment(self.config)
        self.assertFalse(self.create_lxc_base_image_container.called)
        self.assertFalse(self.change_lxc_machine_status.called)
        self.assertFalse(self.configure_lxc_base_machine.called)
        self.assertFalse(self.create_lxc_image_from_container.called)
        self.assertFalse(self.destroy_lxc_machine.called)
        self.logger.debug.has_calls(call("Base image {} found".format(settings.LXC_BASE_IMAGE_ALIAS)))

    def test_ensure_vnet_lxc_environment_calls_base_image_creation_funtions_when_it_does_not_exist(self):
        self.check_if_lxc_image_exists.return_value = False
        ensure_vnet_lxc_environment(self.config)
        self.create_lxc_base_image_container.assert_called_once_with()
        self.change_lxc_machine_status.assert_called_once_with(settings.LXC_BASE_IMAGE_MACHINE_NAME, status="start")
        self.configure_lxc_base_machine.assert_called_once_with()
        self.create_lxc_image_from_container.assert_called_once_with(
            settings.LXC_BASE_IMAGE_MACHINE_NAME, alias=settings.LXC_BASE_IMAGE_ALIAS
        )
        self.destroy_lxc_machine.assert_called_once_with(settings.LXC_BASE_IMAGE_MACHINE_NAME, wait=False)


class TestCleanupVNetLXCEnvironment(VNetTestCase):
    def setUp(self) -> None:
        self.confirm = self.set_up_patch("vnet_manager.environment.lxc.request_confirmation")
        self.delete_vnet_lxc_profile = self.set_up_patch("vnet_manager.environment.lxc.delete_vnet_lxc_profile")
        self.delete_lxc_storage_pool = self.set_up_patch("vnet_manager.environment.lxc.delete_lxc_storage_pool")

    def test_cleanup_vnet_lxc_environment_calls_correct_functions(self):
        cleanup_vnet_lxc_environment()
        self.confirm.assert_called_once_with(message="Cleanup will delete the VNet LXC configurations, such as profile and storage pools")
        self.delete_vnet_lxc_profile.assert_called_once_with(settings.LXC_VNET_PROFILE)
        self.delete_lxc_storage_pool.assert_called_once_with(settings.LXC_STORAGE_POOL_NAME)


class TestConfigureLXCBaseMachine(VNetTestCase):
    def setUp(self) -> None:
        self.get_lxd_client = self.set_up_patch("vnet_manager.environment.lxc.get_lxd_client")
        self.client = MagicMock()
        self.machine = MagicMock()
        self.get_lxd_client.return_value = self.client
        self.client.containers.get.return_value = self.machine
        self.machine.execute.return_value = [0]
        self.sleep = self.set_up_patch("vnet_manager.environment.lxc.sleep")

    def test_configure_lxc_base_machine_check_for_dns(self):
        configure_lxc_base_machine()
        self.machine.execute.assert_has_calls([call(shlex.split("host -t A google.com"))])

    def test_configure_lxc_base_machine_does_not_continue_if_no_dns_connectivity(self):
        self.machine.execute.return_value = [1]
        with self.assertRaises(RuntimeError):
            configure_lxc_base_machine()
        self.assertTrue(self.sleep.called)

    def test_configure_lxc_base_machine_stops_base_if_no_dns_connectivity(self):
        self.machine.execute.return_value = [1]
        with self.assertRaises(RuntimeError):
            configure_lxc_base_machine()
        self.machine.stop.assert_called_once_with()

    def test_configure_lxc_base_machine_calls_correct_configure_cmds(self):
        calls = [
            call(shlex.split("bash -c 'curl -s https://deb.frrouting.org/frr/keys.asc | apt-key add'")),
            call(
                shlex.split(
                    "bash -c 'echo deb https://deb.frrouting.org/frr $(lsb_release -s -c) {} | "
                    "tee -a /etc/apt/sources.list.d/frr.list'".format(settings.FRR_RELEASE)
                )
            ),
            call(shlex.split("apt-get update")),
            call(
                shlex.split("apt-get upgrade -y -o Dpkg::Options::='--force-confdef' -o Dpkg::Options::='--force-confold'"),
                environment={"DEBIAN_FRONTEND": "noninteractive"},
            ),
            call(
                shlex.split(
                    "apt-get install -y -o Dpkg::Options::='--force-confdef' -o Dpkg::Options::='--force-confold' {}".format(
                        " ".join(settings["PROVIDERS"]["lxc"]["guest_packages"])
                    )
                ),
                environment={"DEBIAN_FRONTEND": "noninteractive"},
            ),
            call(shlex.split("systemctl disable radvd")),
            call(shlex.split("bash -c 'echo network: {config: disabled} > /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg'")),
            call(shlex.split("bash -c 'export VTYSH_PAGER=more >> ~/.bashrc'")),
            call(
                shlex.split(
                    'bash -c \'echo -e "#!/bin/bash\nchown -R frr:frr /etc/frr\nsystemctl restart frr" > /etc/rc.local; chmod +x '
                    "/etc/rc.local' "
                )
            ),
        ]
        configure_lxc_base_machine()
        self.machine.execute.assert_has_calls(calls)

    def test_configure_lxc_base_machine_call_machine_stop_(self):
        configure_lxc_base_machine()
        self.machine.stop.assert_called_once_with(wait=True)



from copy import deepcopy
from logging import Formatter, INFO, DEBUG, WARNING, ERROR, CRITICAL
import warnings

try:
    import colorama
except ImportError:
    colorama = None

if colorama:
    default_palette = {
        "asctime": {"*": [colorama.Fore.BLACK, colorama.Style.BRIGHT],},
        "level": {
            DEBUG: [colorama.Fore.WHITE, colorama.Style.DIM],
            WARNING: [colorama.Fore.YELLOW],
            ERROR: [colorama.Fore.RED],
            CRITICAL: [colorama.Fore.RED, colorama.Style.BRIGHT],
        },
        "msg": {
            DEBUG: [colorama.Fore.WHITE, colorama.Style.DIM],
            WARNING: [colorama.Style.BRIGHT],
            ERROR: [colorama.Fore.RED],
            CRITICAL: [colorama.Fore.RED, colorama.Style.BRIGHT],
        },
    }
    _RESET = colorama.Style.RESET_ALL
else:
    default_palette = {"asctime": {}, "level": {}, "msg": {}}
    _RESET = ""


class ConsoleFormatter(Formatter):
    """
    ConsoleFormatter is a log formatter to color log attributes for console printing

    :param str fmt: format string
    :param str datefmt: format string for log date time
    :param bool|callable colored: if should add ANSI colors to logs. If not a bool, it's called as a function
    and the return value is treated as bool
    See logging.Formatter
    """

    def __init__(self, fmt=None, datefmt=None, colored=True):
        Formatter.__init__(self, fmt, datefmt)
        self._colored = bool(colored()) if colored and not isinstance(colored, bool) else colored
        if self._colored and not colorama:
            warnings.warn("can't format colored log message. dependency package 'colorama' is not installed")
            self._colored = False
        self._palette = default_palette if self._colored else {"asctime": {}, "level": {}, "msg": {}}

    @property
    def colored(self):
        return self._colored

    def format(self, record):
        """
        Format the specified record as text.

        The record's attribute dictionary is used as the operand to a
        string formatting operation which yields the returned string.
        Before formatting the dictionary, a couple of preparatory steps
        are carried out. The message attribute of the record is computed
        using LogRecord.getMessage(). If the formatting string uses the
        time (as determined by a call to usesTime(), formatTime() is
        called to format the event time. If there is exception information,
        it is formatted using formatException() and appended to the message.

        See logging.Format.format
        """
        if self._colored:
            return self._colored_format(record)
        return Formatter.format(self, record)

    def _colored_format(self, record):
        """
        Format the specified record as text colored using ANSI escape sequences

        See format
        """
        record = deepcopy(record)  # avoid mutating the record itself
        level = getattr(record, "levelno", INFO)

        level_palette = self._palette["level"]
        level_styles = level_palette[level] if level in level_palette else level_palette.get("*", [])
        if level_styles:
            record.levelname = r"".join(level_styles) + str(record.levelname) + _RESET

        msg_palette = self._palette["msg"]
        msg_styles = msg_palette[level] if level in msg_palette else msg_palette.get("*", [])
        if msg_styles:
            record.msg = r"".join(msg_styles) + str(record.msg) + _RESET

        return Formatter.format(self, record)

    def formatTime(self, record, datefmt=None):
        """
        Return the creation time of the specified LogRecord as formatted text.

        This method should be called from format()

        See logging.Formatter.formatTime
        """
        formatted_time = Formatter.formatTime(self, record, datefmt)
        if self._colored:
            level = getattr(record, "levelno", INFO)
            asctime_palette = self._palette["asctime"]
            asctime_styles = asctime_palette[level] if level in asctime_palette else asctime_palette.get("*", [])
            if asctime_styles:
                return r"".join(asctime_styles) + formatted_time + _RESET
        return formatted_time



import shlex
from logging import getLogger
from time import sleep

from vnet_manager.operations.image import check_if_lxc_image_exists, create_lxc_image_from_container
from vnet_manager.operations.profile import check_if_lxc_profile_exists, create_vnet_lxc_profile, delete_vnet_lxc_profile
from vnet_manager.operations.storage import check_if_lxc_storage_pool_exists, create_lxc_storage_pool, delete_lxc_storage_pool
from vnet_manager.operations.machine import create_lxc_base_image_container, change_lxc_machine_status, destroy_lxc_machine
from vnet_manager.environment.host import check_for_supported_os, check_for_installed_packages
from vnet_manager.providers.lxc import get_lxd_client
from vnet_manager.conf import settings
from vnet_manager.utils.user import request_confirmation

logger = getLogger(__name__)


def ensure_vnet_lxc_environment(config):
    """
    Checks and creates the LXC environment
    param: dict config: The config created by get_config()
    :raises RuntimeError: If unsupported OS, or missing packages
    """
    # Check if there are any LXC machines in the config
    if "lxc" not in [settings.MACHINE_TYPE_PROVIDER_MAPPING[machine["type"]] for machine in config["machines"].values()]:
        logger.debug("Skipping LXC environment creation, no LXC machines in config")
        return

    # Check if we are on a supported OS
    if not check_for_supported_os("lxc"):
        logger.critical("Unable to create LXC environment on your machine, OS not supported")
        raise RuntimeError("OS not supported for provider LXC")

    # Check if all required packages have been installed
    if not check_for_installed_packages("lxc"):
        logger.critical("Not all required host packages seem to be installed, please fix this before proceeding")
        raise RuntimeError("Missing host packages")

    # Check if the storage pool exists
    if not check_if_lxc_storage_pool_exists(settings.LXC_STORAGE_POOL_NAME):
        logger.info("VNet LXC storage pool does not exist, creating it")
        create_lxc_storage_pool(name=settings.LXC_STORAGE_POOL_NAME, driver=settings.LXC_STORAGE_POOL_DRIVER)
    else:
        logger.debug("VNet LXC storage pool {} found".format(settings.LXC_STORAGE_POOL_NAME))

    # Check if the profile exists
    if not check_if_lxc_profile_exists(settings.LXC_VNET_PROFILE):
        logger.info("VNet LXC profile does not exist, creating it")
        create_vnet_lxc_profile(settings.LXC_VNET_PROFILE)
    else:
        logger.debug("VNet profile {} found".format(settings.LXC_VNET_PROFILE))

    # Check if the base image exists
    if not check_if_lxc_image_exists(settings.LXC_BASE_IMAGE_ALIAS, by_alias=True):
        logger.info("Base image does not exist, creating it")
        create_lxc_base_image_container()
        change_lxc_machine_status(settings.LXC_BASE_IMAGE_MACHINE_NAME, status="start")
        configure_lxc_base_machine()
        create_lxc_image_from_container(settings.LXC_BASE_IMAGE_MACHINE_NAME, alias=settings.LXC_BASE_IMAGE_ALIAS)
        destroy_lxc_machine(settings.LXC_BASE_IMAGE_MACHINE_NAME, wait=False)
    else:
        logger.debug("Base image {} found".format(settings.LXC_BASE_IMAGE_ALIAS))


def cleanup_vnet_lxc_environment():
    """
    Cleans up specific VNet LXC configuration
    No environments should be active when calling this function
    """
    request_confirmation(message="Cleanup will delete the VNet LXC configurations, such as profile and storage pools")
    logger.info("Cleaning up VNet LXC configuration")
    delete_vnet_lxc_profile(settings.LXC_VNET_PROFILE)
    delete_lxc_storage_pool(settings.LXC_STORAGE_POOL_NAME)


def configure_lxc_base_machine():
    """
    Configure the LXC base machine to get a fully functional VNet base machine which we can make an image from
    :raises RuntimeError: If the base machine is started without networking/dns
    """
    logger.info("Configuring LXC base machine {}, this might take a while".format(settings.LXC_BASE_IMAGE_MACHINE_NAME))
    client = get_lxd_client()
    machine = client.containers.get(settings.LXC_BASE_IMAGE_MACHINE_NAME)

    def execute_and_log(command, **kwargs):
        result = machine.execute(shlex.split(command), **kwargs)
        logger.debug(result)
        return result

    # Check for DNS
    logger.debug("Checking for DNS connectivity")
    dns = False
    for _ in range(0, settings.LXC_MAX_STATUS_WAIT_ATTEMPTS):
        if execute_and_log("host -t A google.com")[0] == 0:
            dns = True
            break
        # No DNS connectivity (yet), try again
        sleep(2)
    if not dns:
        # Shutdown base if DNS check fails
        logger.debug("Stopping base machine")
        machine.stop()
        raise RuntimeError("Base machine started without working DNS, unable to continue")

    # Set the FRR routing source and key
    execute_and_log("bash -c 'curl -s https://deb.frrouting.org/frr/keys.asc | apt-key add'")
    execute_and_log(
        "bash -c 'echo deb https://deb.frrouting.org/frr $(lsb_release -s -c) {} | tee -a /etc/apt/sources.list.d/frr.list'".format(
            settings.FRR_RELEASE
        )
    )

    # Update and install packages
    execute_and_log("apt-get update")
    execute_and_log(
        "apt-get upgrade -y -o Dpkg::Options::='--force-confdef' -o Dpkg::Options::='--force-confold'",
        environment={"DEBIAN_FRONTEND": "noninteractive"},
    )
    execute_and_log(
        "apt-get install -y -o Dpkg::Options::='--force-confdef' -o Dpkg::Options::='--force-confold' {}".format(
            " ".join(settings["PROVIDERS"]["lxc"]["guest_packages"])
        ),
        environment={"DEBIAN_FRONTEND": "noninteractive"},
    )

    # Disable radvd by default
    execute_and_log("systemctl disable radvd")
    # Disable cloud init messing with our networking
    execute_and_log("bash -c 'echo network: {config: disabled} > /etc/cloud/cloud.cfg.d/99-disable-network-config.cfg'")
    # Set the default VTYSH_PAGER
    execute_and_log("bash -c 'export VTYSH_PAGER=more >> ~/.bashrc'")
    # Make all files in the FRR dir owned by the frr user
    execute_and_log(
        "bash -c 'echo -e \"#!/bin/bash\nchown -R frr:frr /etc/frr\nsystemctl restart frr\" > /etc/rc.local; chmod +x /etc/rc.local'"
    )
    # All done, stop the container
    machine.stop(wait=True)
    logger.debug("LXC base machine {} successfully configured".format(settings.LXC_BASE_IMAGE_MACHINE_NAME))



from logging import getLogger
from pylxd.exceptions import NotFound
from os.path import isfile, isdir, join, basename
from os import listdir
from sys import modules

from vnet_manager.providers.lxc import get_lxd_client
from vnet_manager.conf import settings
from vnet_manager.utils.files import write_file_to_disk

logger = getLogger(__name__)


def put_files_on_machine(config):
    for name, data in config["machines"].items():
        if "files" in data:
            provider = settings.MACHINE_TYPE_PROVIDER_MAPPING[data["type"]]
            logger.info("Putting requested files on machine {}".format(name))
            select_files_and_put_on_machine(name, data["files"], provider)


def select_files_and_put_on_machine(machine, files, provider):
    """
    Checks if the requested files are files or a dict and calls the provider file place function for each file
    :param str machine: The machine to put the files on
    :param dict files: The machines file dict from the config
    :param str provider: The provider of the machine
    """
    # Get the files
    for host_path, guest_path in files.items():
        if isdir(host_path):
            logger.debug("Getting files from file dir {}".format(host_path))
            files = [join(host_path, f) for f in listdir(host_path)]
            for file_path in files:
                # Place the file on the machine
                getattr(modules[__name__], "place_file_on_{}_machine".format(provider))(
                    machine, file_path, join(guest_path, basename(file_path))
                )
        elif isfile(host_path):
            getattr(modules[__name__], "place_file_on_{}_machine".format(provider))(machine, host_path, guest_path)
        else:
            logger.error("Tried to select file {} for copying, but it is neither a file nor a directory".format(host_path))


def write_file_to_lxc_container(container, file_path, data):
    """
    Writes file data to a path on the LXC container
    :param container: str: The name of the container
    :param file_path: str: The guest path to write the file to
    :param data: The data to write
    """
    try:
        machine = get_lxd_client().containers.get(container)
        machine.files.put(file_path, data)
    except NotFound:
        logger.error("Tried to write data to path {} on LXC container {}, but the container does not exist".format(file_path, container))


def place_file_on_lxc_machine(container, host_file_path, guest_file_path):
    """
    Places a local file on a LXC container
    :param str container: The container to place the file on
    :param str host_file_path: The file to copy to the guest
    :param str guest_file_path: The path to copy the file to
    """
    # Some sanity checks
    if not isfile(host_file_path):
        logger.error("Tried to copy {} to LXC container {}, but the file doesn't exists".format(host_file_path, container))
        return

    # Get the file contents
    with open(host_file_path, "r") as fh:
        file_data = fh.read()
    # Place the file content
    logger.debug("Copying {} to container {} at path {}".format(host_file_path, container, guest_file_path))
    write_file_to_lxc_container(container, guest_file_path, file_data)


def generate_vnet_hosts_file(config):
    """
    Generates the machines /etc/hosts file based on the info in the config
    The generated file is placed on disk
    :param dict config: The config generated by get_config()
    """
    logger.info("Generating VNet hosts file")
    vnet_hosts = []
    for machine_name, machine_data in config["machines"].items():
        for int_data in machine_data["interfaces"].values():
            if "ipv4" in int_data:
                vnet_hosts.append("{}   {}".format(int_data["ipv4"].split("/")[0], machine_name))
            if "ipv6" in int_data:
                vnet_hosts.append("{}   {}".format(int_data["ipv6"].split("/")[0], machine_name))
    vnet_etc_hosts_data = settings.VNET_STATIC_HOSTS_FILE_PART + "\n".join(vnet_hosts)
    write_file_to_disk(settings.VNET_ETC_HOSTS_FILE_PATH, vnet_etc_hosts_data)


def place_vnet_hosts_file_on_machines(config):
    """
    Places the generated /etc/hosts file on all VNet machines defined in the config
    :param dict config: The config generated by get_config()
    """
    logger.info("Placing VNet /etc/hosts file on machines")
    for name, data in config["machines"].items():
        provider = settings.MACHINE_TYPE_PROVIDER_MAPPING[data["type"]]
        select_files_and_put_on_machine(name, {settings.VNET_ETC_HOSTS_FILE_PATH: "/etc/hosts"}, provider)



from unittest.mock import MagicMock

from vnet_manager.tests import VNetTestCase
from vnet_manager.environment.host import check_for_installed_packages, check_for_supported_os
from vnet_manager.conf import settings


class TestCheckForSupportedOS(VNetTestCase):
    def setUp(self) -> None:
        self.codename = self.set_up_patch("vnet_manager.environment.host.codename")
        self.codename.return_value = "bionic"

    def test_check_for_supported_os_returns_true_if_os_supported(self):
        self.assertTrue(check_for_supported_os("lxc"))

    def test_check_for_supported_os_returns_false_if_os_not_supported(self):
        self.codename.return_value = "Hannah Montana Linux"
        self.assertFalse(check_for_supported_os("lxc"))


class TestCheckForInstalledPackages(VNetTestCase):
    def setUp(self) -> None:
        self.cache = self.set_up_patch("vnet_manager.environment.host.Cache", themock=MagicMock())
        self.logger = self.set_up_patch("vnet_manager.environment.host.logger")

    def test_check_for_installed_packages_calls_apt_cache(self):
        check_for_installed_packages("lxc")
        self.cache.assert_called_once_with()

    def test_check_for_installed_packages_returns_true_if_all_packages_installed(self):
        self.assertTrue(check_for_installed_packages("lxc"))

    def test_check_for_installed_packages_returns_false_if_not_all_packages_installed(self):
        self.cache.return_value[settings["PROVIDERS"]["lxc"]["required_host_packages"][0]].is_installed = False
        self.assertFalse(check_for_installed_packages("lxc"))



from pkg_resources import require

from vnet_manager.conf import settings


def show_version():
    print(
        """
VNet manager version {}
""".format(
            require(settings.PYTHON_PACKAGE_NAME)[0].version
        )
    )



from unittest.mock import call

from vnet_manager.conf import import_from_string
from vnet_manager.tests import VNetTestCase


class TestImportFromString(VNetTestCase):
    def setUp(self) -> None:
        self.mock_import_module = self.set_up_patch("importlib.import_module", return_value="module")

        self.mock_getattr = self.set_up_patch("vnet_manager.conf.getattr", return_value="attr")

    def test_that_method_calls_import_module_correct(self):
        import_from_string("module.class.sub", "xyz")
        expected_calls = [call("module.class")]
        self.assertEqual(expected_calls, self.mock_import_module.mock_calls)

    def test_that_method_calls_get_attr_correct(self):
        import_from_string("module.class.sub", "xyz")
        expected_calls = [call("module", "sub")]
        self.assertEqual(expected_calls, self.mock_getattr.mock_calls)

    def test_that_method_raises_importerror_if_import_module_raises_importerror(self):
        self.mock_import_module.side_effect = ImportError
        with self.assertRaises(ImportError):
            import_from_string("module.class.sub", "xyz")

    def test_that_method_raises_importerror_if_getattr_raises_attributerror(self):
        self.mock_getattr.side_effect = AttributeError
        with self.assertRaises(ImportError):
            import_from_string("module.class.sub", "xyz")



import logging
import logging.config

from vnet_manager.conf import settings


def setup_console_logging(verbosity=logging.INFO):
    """
    :param int verbosity: Verbosity level logging.<verbosity>
    """
    settings.LOGGING["handlers"]["console"]["level"] = verbosity
    settings.LOGGING["handlers"]["syslog"]["level"] = verbosity
    logging.config.dictConfig(settings.LOGGING)



from logging import getLogger
from os.path import realpath, dirname

from vnet_manager.utils.files import get_yaml_content

logger = getLogger(__name__)


def get_config(path):
    """
    Produces the vnet config that can be used by other functions
    :param str path: The path to load the user config from (overwrites the default config)
    :return: dict: The merged user and default configs
    """
    logger.debug("Getting user config")
    config = get_yaml_content(path)
    # Add the config directory
    config["config_dir"] = dirname(realpath(path))
    return config



from pylxd.exceptions import NotFound, LXDAPIException
from sys import modules
from logging import getLogger
from tabulate import tabulate
from time import sleep
from yaml import safe_dump

from vnet_manager.conf import settings
from vnet_manager.operations.files import write_file_to_lxc_container
from vnet_manager.providers.lxc import get_lxd_client
from vnet_manager.utils.user import request_confirmation

logger = getLogger(__name__)


def show_status(config):
    """
    Print a table with the current machine statuses
    :param dict config: The config provided by vnet_manager.config.get_config()
    """
    logger.info("Listing VNet machine statuses")
    header = ["Name", "Status", "Provider"]
    statuses = []
    for name, info in config["machines"].items():
        provider = settings.MACHINE_TYPE_PROVIDER_MAPPING[info["type"]]
        # Call the relevant provider get_%s_machine_status function
        statuses.append(getattr(modules[__name__], "get_{}_machine_status".format(provider))(name))
    print(tabulate(statuses, headers=header, tablefmt="pretty"))


def check_if_lxc_machine_exists(machine):
    """
    Checks if an LXC machine exists
    :param str machine: The machine/container to check for
    :return: bool: True if it exists, false otherwise
    """
    return get_lxd_client().containers.exists(machine)


def get_lxc_machine_status(name):
    """
    Gets the LXC machine state and returns a list
    :param name: str: The name of the machine
    :return: list: [name, state, provider]
    """
    # TODO: Let's not return a list here, simply the status
    client = get_lxd_client()
    try:
        status = client.containers.get(name).status
    except NotFound:
        status = "NA"
    return [name, status, "LXC"]


def wait_for_lxc_machine_status(container, status):
    """
    Waits for a LXC machine to converge to the requested status
    :param pylxd.client.Client.container() container: The container to wait for
    :param str status: The status to wait for
    :raise TimeoutError if wait time expires
    """
    logger.debug("Waiting for LXC container {} to get a {} status".format(container.name, status))
    for i in range(1, settings.LXC_MAX_STATUS_WAIT_ATTEMPTS):
        # Actually ask for the container.state().status, because container.status is a static value
        if container.state().status.lower() == status.lower():
            logger.debug("Container successfully converged to {} status".format(status))
            return
        # Container not in desired state yet, wait and try again
        sleep_time = i * (settings.LXC_STATUS_WAIT_SLEEP * settings.LXC_STATUS_BACKOFF_MULTIPLIER)
        logger.info("Container {} not yet in {} status, waiting for {} seconds".format(container.name, status, sleep_time))
        sleep(sleep_time)
    raise TimeoutError("Wait time for container {} to converge to {} status expired, giving up".format(container.name, status))


def change_machine_status(config, status="stop", machines=None):
    """
    Change the status of the passed machines to the requested state
    :param dict config: The config provided by get_config()
    :param str status: The status to change the machine to
    :param list machines: A list of machine names to stop/start, if None all will be changed
    """
    # Check for valid status change
    if status not in settings.VALID_STATUSES:
        raise NotImplementedError("Requested machine status change {} unknown".format(status))

    # Get all the machines from the config if not already provided
    machines = machines if machines else config["machines"].keys()

    # For each machine get the provider and execute the relevant status change function
    for machine in machines:
        # First check if the machine exists
        if machine not in config["machines"]:
            logger.error("Tried to {} machine {}, but there is no config entry for it, skipping...".format(status, machine))
            continue
        # Get the provider
        provider = settings.MACHINE_TYPE_PROVIDER_MAPPING[config["machines"][machine]["type"]]
        # Call the provider change_status function
        logger.info("{} machine {} with provider {}".format("Starting" if status == "start" else "Stopping", machine, provider))
        getattr(modules[__name__], "change_{}_machine_status".format(provider))(machine, status=status)


def change_lxc_machine_status(machine, status="stop"):
    """
    Start a LXC machine
    :param str machine: The name of the machine to change the status of
    :param str status: The status to change the LXC machine to
    """
    client = get_lxd_client()
    try:
        machine = client.containers.get(machine)
    except NotFound:
        logger.error("Tried to change machine status of LXC container {}, but it doesn't exist!".format(machine))
        return
    # Change the status
    if status == "stop":
        machine.stop()
    elif status == "start":
        try:
            # On start we wait, as we might catch invalid configs
            machine.start(wait=True)
        except LXDAPIException as e:
            logger.error("Unable to start LXC container {}, got error: {}".format(machine.name, e))
            return
    # Take a short nap after issuing the start/stop command, so we might pass the first status check
    sleep(1)
    try:
        required_state = "Stopped" if status == "stop" else "Running"
        wait_for_lxc_machine_status(machine, required_state)
        logger.debug("LXC container {} is {}".format(machine.name, machine.state().status))
    except TimeoutError:
        logger.error("Unable to change LXC status container {}, got timeout after issuing {} command".format(machine.name, status))


def create_machines(config, machines=None):
    """
    Meta function to call the other machine creation functions per provider
    :param dict config: The config generated by get_config()
    :param list machines: A list of machine to create, defaults to all machines in the config
    """
    # Get all the machines from the config if not already provided
    machines = machines if machines else config["machines"].keys()
    create_lxc_machines_from_base_image(config, machines)


def create_lxc_machines_from_base_image(config, containers):
    """
    Create LXC machines from the base image specified in the settings
    :param dict config: The config generated by get_config()
    :param list containers: A list of machines to create
    """
    containers_to_create = []
    containers_already_created = False

    # Get all the LXC machines to create
    for container in containers:
        # Check if the requested machine name is present in the config
        if container not in config["machines"]:
            logger.error(
                "Tried to get provider for container {}, but the container was not found in the config, skipping".format(container)
            )
        # Quick check if the machine already exists
        elif check_if_lxc_machine_exists(container):
            logger.error("A LXC container with the name {} already exists, skipping".format(container))
            containers_already_created = True
        # Check if LXC is the provider
        elif settings.MACHINE_TYPE_PROVIDER_MAPPING[config["machines"][container]["type"]].lower() == "lxc":
            logger.debug("Selecting LXC machine {} for creation".format(container))
            containers_to_create.append(container)
        else:
            logger.debug("Machine {} is not provided by LXC, skipping LXC container creation".format(container))

    # Create it
    client = get_lxd_client()
    for container in containers_to_create:
        logger.debug("Generating LXC config for container {}".format(container))
        # Interface config
        # First add eth0 (default), which does nothing
        device_config = {"eth0": {"type": "none"}}
        # Then for each interface in the config add the configuration for that interface to the interfaces_config dict
        for inet_name, inet_config in config["machines"][container]["interfaces"].items():
            device_config[inet_name] = {
                "name": inet_name,  # The name of the interface inside the instance
                "host_name": "{}-{}".format(container, inet_name),  # The name of the interface inside the host
                "parent": "{}{}".format(settings.VNET_BRIDGE_NAME, inet_config["bridge"]),  # The name of the host device
                "type": "nic",
                "nictype": "bridged",
                "hwaddr": inet_config["mac"],
            }
        container_config = {
            "name": container,
            "source": {"alias": settings.LXC_BASE_IMAGE_ALIAS, "type": "image"},
            "ephemeral": False,
            "config": {"user.network-config": "disabled"},
            "devices": device_config,
            "profiles": [settings.LXC_VNET_PROFILE],
        }
        logger.info("Creating LXC container {}".format(container))
        # TODO: Make this nicer by not waiting here but doing the configuration after we've created all containers
        client.containers.create(container_config, wait=True)
        place_lxc_interface_configuration_on_container(config, container)

    # Check with the user if it is okay to overwrite config files
    if containers_already_created:
        request_confirmation(
            message="Some containers already existed, the next operation will overwrite network, "
            "host and user config files on those containers",
        )


def destroy_machines(config, machines=None):
    """
    Destroy's the passed machines
    :param dict config: The config generated by get config
    :param list machines: The machines to destroy, defaults to all machines in the config
    :param bool force: Don't ask the user questions, just do it
    """
    # Get all the machines from the config if not already provided
    machines = machines if machines else config["machines"].keys()

    # Ask the user if he is sure
    request_confirmation(
        message="Requesting confirmation of deletion for the following machines: {}".format(", ".join(machines)),
        prompt="This operation cannot be undone. Are you sure?! (yes/no) ",
    )

    for machine in machines:
        # First check if the machine exists
        if machine not in config["machines"]:
            logger.error("Tried to get the config for machine {}, but there is no config entry for this machine, skipping".format(machine))
            continue
        # Get the provider
        provider = settings.MACHINE_TYPE_PROVIDER_MAPPING[config["machines"][machine]["type"]]
        # Call the provider destroy function
        getattr(modules[__name__], "destroy_{}_machine".format(provider))(machine)


def destroy_lxc_machine(machine, wait=False):
    """
    Deletes an LXC machine
    :param str machine: The name of the machine to delete
    :param bool wait: Wait for the deletion to be complete before returning
    :return:
    """
    client = get_lxd_client()
    try:
        container = client.containers.get(machine)
    except NotFound:
        logger.warning("Tried to delete LXC machine {}, but it does not exist. Maybe it was already deleted?".format(machine))
        return
    # Check if the container is still running
    if container.status.lower() == "running":
        logger.info("Stopping LXC container {}".format(machine))
        container.stop(wait=True)
    logger.info("Deleting LXC container {}".format(machine))
    container.delete(wait=wait)


def create_lxc_base_image_container():
    """
    Creates the LXC base image container
    :raises RuntimeError: if any issues are encountered during creation
    """
    # Check if the base image machine already exists and destroy it
    if check_if_lxc_machine_exists(settings.LXC_BASE_IMAGE_MACHINE_NAME):
        logger.warning("LXC base image machine already exists")
        request_confirmation(message="Recreating it will destroy any local changes", prompt="Recreate the LXC base image machine? ")
        destroy_lxc_machine(settings.LXC_BASE_IMAGE_MACHINE_NAME, wait=True)

    # Now create the base image machine
    client = get_lxd_client()

    machine_config = {
        "name": settings.LXC_BASE_IMAGE_MACHINE_NAME,
        "architecture": "x86_64",
        "profiles": [settings.LXC_VNET_PROFILE],
        "ephemeral": False,
        "config": {},
        "devices": {
            "eth0": {
                "name": "eth0",
                "parent": "lxdbr0",
                "type": "nic",
                "nictype": "bridged",
                "host_name": "{}-eth0".format(settings.LXC_BASE_IMAGE_MACHINE_NAME),
            }
        },
        "source": {
            "type": "image",
            "protocol": str(settings["PROVIDERS"]["lxc"]["base_image"]["protocol"]),
            "server": settings["PROVIDERS"]["lxc"]["base_image"]["server"],
            "alias": str(settings["PROVIDERS"]["lxc"]["base_image"]["os"]),
        },
    }
    logger.info("Creating LXC base image container")
    client.containers.create(machine_config, wait=True)


def enable_type_specific_machine_configuration(config):
    """
    Call type and provider specific machine configuration functions based on the settings
    :param dict config: The config generated by get_config()
    """
    for machine_name, machine_data in config["machines"].items():
        for func in settings.MACHINE_TYPE_CONFIG_FUNCTION_MAPPING[machine_data["type"]]:
            getattr(modules[__name__], func)(machine_name)


def enable_lxc_ip_forwarding(container_name):
    """
    Enables LXC IP forwarding for a machine
    (Wrapper function for configure_lxc_ip_forwarding)
    :param container_name: str The name of the container
    """
    configure_lxc_ip_forwarding(container_name, enable=True)


def disable_lxc_ip_forwarding(container_name):
    """
    Disables LXC IP forwarding for a machine
    (Wrapper function for configure_lxc_ip_forwarding)
    :param container_name: str The name of the container
    """
    configure_lxc_ip_forwarding(container_name, enable=False)


def configure_lxc_ip_forwarding(container_name, enable=True):
    """
    Configure a LXC machine to enable IP forwarding
    :param str container_name: The name of the container to enable IP forwarding on
    :param bool enable: Whether to enable IP forwarding or disable it
    """
    value = 1 if enable else 0
    logger.info("{} IP forwarding on LXC container {}".format("Enabling" if enable else "Disabling", container_name))
    write_file_to_lxc_container(container_name, "/etc/sysctl.d/20-net.ipv4.ip_forward.conf", "net.ipv4.ip_forward={}\n".format(value))
    write_file_to_lxc_container(
        container_name, "/etc/sysctl.d/20-net.ipv6.conf.all.forwarding.conf", "net.ipv6.conf.all.forwarding={}\n".format(value)
    )


def place_lxc_interface_configuration_on_container(config, container):
    """
    Places the interfaces configuration on the LXC container
    :param dict config: The config generated by get_config()
    :param str container: The name of the container to place the interfaces configuration on
    """
    logger.debug("Generating network config for LXC container {}".format(container))
    network_conf = generate_machine_netplan_config(config, container)
    logger.info("Placing network config on LXC container {}".format(container))
    write_file_to_lxc_container(container, settings.VNET_NETPLAN_CONFIG_FILE_PATH, safe_dump(network_conf))


def generate_machine_netplan_config(config, machine):
    """
    Generates a Netplan config based on the machine configuration
    :param config: dict: The config generates by get_config()
    :param machine: str: The machine name to generate the netplan config for
    :return: dict: The netplan config
    """
    no_dhcp = {
        "dhcp4": "no",
        "dhcp6": "no",
    }
    machine_config = config.get("machines").get(machine)
    network_conf = {"network": {"version": 2, "renderer": "networkd", "ethernets": {},}}
    for int_name, int_data in machine_config["interfaces"].items():
        addresses = []
        if "ipv4" in int_data:
            addresses.append(int_data["ipv4"])
        if "ipv6" in int_data:
            addresses.append(int_data["ipv6"])
        network_conf["network"]["ethernets"][int_name] = {
            "match": {"macaddress": int_data["mac"],},
            "set-name": int_name,
            "addresses": addresses,
        }
        network_conf["network"]["ethernets"][int_name].update(no_dhcp)
    if "vlans" in machine_config:
        network_conf["network"]["vlans"] = {}
        for vlan, vlan_data in machine_config["vlans"].items():
            network_conf["network"]["vlans"][vlan] = {"id": vlan_data["id"], "link": vlan_data["link"], "addresses": vlan_data["addresses"]}
            network_conf["network"]["vlans"][vlan].update(no_dhcp)
    return network_conf



from unittest.mock import MagicMock
from pylxd.exceptions import NotFound

from vnet_manager.tests import VNetTestCase
from vnet_manager.operations.image import check_if_lxc_image_exists, create_lxc_image_from_container, destroy_lxc_image


class TestCheckIfLXCImageExists(VNetTestCase):
    def setUp(self) -> None:
        self.client = self.set_up_patch("vnet_manager.operations.image.get_lxd_client")
        self.images = MagicMock()
        self.client.return_value.images = self.images
        self.images.get_by_alias.return_value = True
        self.images.exists.return_value = True

    def test_check_if_lxc_image_exists_calls_get_lxd_client(self):
        check_if_lxc_image_exists("test")
        self.client.assert_called_once_with()

    def test_check_if_lxc_image_exists_call_get_by_alias(self):
        check_if_lxc_image_exists("test")
        self.images.get_by_alias.assert_called_once_with("test")

    def test_check_if_lxc_image_exists_returns_true_if_get_by_alias_returns_true(self):
        self.images.exists.return_value = False
        self.assertTrue(check_if_lxc_image_exists("test"))

    def test_check_if_lxc_image_exists_does_not_call_exists_if_get_by_alias_succeeds(self):
        self.assertTrue(check_if_lxc_image_exists("test"))
        self.assertFalse(self.images.exists.called)

    def test_check_if_lxc_image_exists_does_not_call_exists_if_get_by_alias_fails(self):
        self.images.get_by_alias.side_effect = NotFound(response="blaap")
        check_if_lxc_image_exists("test")
        self.assertFalse(self.images.exists.called)

    def test_check_if_lxc_image_exists_returns_false_if_get_by_alias_raises_not_found(self):
        self.images.get_by_alias.side_effect = NotFound(response="blaap")
        self.assertFalse(check_if_lxc_image_exists("test"))

    def test_check_if_lxc_image_exists_calls_exists_when_not_checking_by_alias(self):
        check_if_lxc_image_exists("test", by_alias=False)
        self.images.exists.assert_called_once_with("test")
        self.assertFalse(self.images.get_by_alias.called)

    def test_check_if_lxc_image_exists_returns_true_when_exists_returns_true(self):
        self.assertTrue(check_if_lxc_image_exists("test", by_alias=False))

    def test_check_if_lxc_image_returns_false_when_exists_returns_false(self):
        self.images.exists.return_value = False
        self.assertFalse(check_if_lxc_image_exists("test", by_alias=False))


class TestCreateLXCImageFromContainer(VNetTestCase):
    def setUp(self) -> None:
        self.change_status = self.set_up_patch("vnet_manager.operations.image.change_lxc_machine_status")
        self.lxd_client = self.set_up_patch("vnet_manager.operations.image.get_lxd_client")
        self.client = MagicMock()
        self.lxd_client.return_value = self.client
        self.container = MagicMock()
        self.image = MagicMock()
        self.client.containers.get.return_value = self.container
        self.container.publish.return_value = self.image

    def test_create_lxc_image_from_container_call_change_lxc_machine_status(self):
        create_lxc_image_from_container("test")
        self.change_status.assert_called_once_with("test", status="stop")

    def test_create_lxc_image_from_container_calls_get_lxd_client(self):
        create_lxc_image_from_container("test")
        self.lxd_client.assert_called_once_with()

    def test_create_lxc_image_from_container_calls_container_get_method(self):
        create_lxc_image_from_container("test")
        self.client.containers.get.assert_called_once_with("test")

    def test_create_lxc_image_from_container_calls_publish_on_container(self):
        create_lxc_image_from_container("test")
        self.container.publish.assert_called_once_with(wait=True)

    def test_create_lxc_image_from_container_does_not_add_alias_by_default(self):
        create_lxc_image_from_container("test")
        self.assertFalse(self.image.called)
        self.assertFalse(self.image.add_alias.called)

    def test_create_lxc_image_from_container_add_alias_when_requested(self):
        create_lxc_image_from_container("test", alias="test_alias", description="test_desc")
        self.image.add_alias.assert_called_once_with("test_alias", "test_desc")


class TestDestroyLXCImage(VNetTestCase):
    def setUp(self) -> None:
        self.check_if_image_exists = self.set_up_patch("vnet_manager.operations.image.check_if_lxc_image_exists")
        self.get_lxd = self.set_up_patch("vnet_manager.operations.image.get_lxd_client")
        self.client = MagicMock()
        self.get_lxd.return_value = self.client
        self.image = MagicMock()
        self.client.images.get_by_alias.return_value = self.image
        self.client.images.get.return_value = self.image

    def test_destroy_lxc_image_checks_if_image_exists(self):
        destroy_lxc_image("test")
        self.check_if_image_exists.assert_called_once_with("test", by_alias=True)

    def test_destroy_lxc_image_check_if_image_exists_without_alias(self):
        destroy_lxc_image("test", by_alias=False)
        self.check_if_image_exists.assert_called_once_with("test", by_alias=False)

    def test_destroy_lxc_image_calls_get_lxd_client(self):
        destroy_lxc_image("test")
        self.get_lxd.assert_called_once_with()

    def test_destroy_lxc_image_calls_images_get_by_alias_by_default(self):
        destroy_lxc_image("test")
        self.client.images.get_by_alias.assert_called_once_with("test")
        self.assertFalse(self.client.images.get.called)

    def test_destroy_lxc_image_calls_images_get_when_not_by_alias(self):
        destroy_lxc_image("test", by_alias=False)
        self.client.images.get.assert_called_once_with("test")
        self.assertFalse(self.client.images.get_by_alias.called)

    def test_destroy_lxc_image_does_nothing_when_image_does_not_exist(self):
        self.check_if_image_exists.return_value = False
        destroy_lxc_image("test")
        self.assertFalse(self.get_lxd.called)
        self.assertFalse(self.image.delete.called)

    def test_destroy_lxc_image_calls_image_delete_method(self):
        destroy_lxc_image("test")
        self.image.delete.assert_called_once_with()



from logging import getLogger
from pylxd.exceptions import NotFound

from vnet_manager.providers.lxc import get_lxd_client
from vnet_manager.operations.machine import change_lxc_machine_status

logger = getLogger(__name__)


def check_if_lxc_image_exists(image, by_alias=True):
    """
    Check if an LXC image exists
    :param str image: The image fingerprint or alias
    :param bool by_alias: Search by image alias instead of fingerprint
    :return: bool: True if the image exists false otherwise
    """
    logger.debug("Checking for the existence of LXC image {}".format(image))
    client = get_lxd_client()
    if by_alias:
        try:
            client.images.get_by_alias(image)
            return True
        except NotFound:
            return False
    # Search by fingerprint
    return client.images.exists(image)


def create_lxc_image_from_container(container, alias=None, description=None):
    """
    Create a LXC image from a container
    :param str container: The container to create the image from
    :param str alias: Creates an alias for the image
    :param str description: A description for the image alias
    """
    # Stop it first
    change_lxc_machine_status(container, status="stop")

    # Create the image
    logger.info("Creating image from LXC container {}".format(container))
    client = get_lxd_client()
    container = client.containers.get(container)
    img = container.publish(wait=True)
    logger.info("Image {} created successfully".format(img.fingerprint))

    # Create the alias if requested
    if alias:
        logger.info("Adding alias {} to newly created image".format(alias))
        img.add_alias(alias, description)


def destroy_lxc_image(image, by_alias=True):
    """
    Destroy a LXC image
    :param str image: The fingerprint or alias of the image to destroy
    :param bool by_alias: Search by alias instead of fingerprint
    """
    # Check if it even exists
    if not check_if_lxc_image_exists(image, by_alias=by_alias):
        logger.warning("Tried to destroy LXC image {}, but it is already gone".format(image))
        return
    # Delete it
    logger.info("Deleting LXC image {}".format(image))
    client = get_lxd_client()
    image = client.images.get_by_alias(image) if by_alias else client.images.get(image)
    image.delete()



from logging import getLogger
from os import EX_OK, EX_USAGE
from os.path import isdir, isfile
from warnings import warn

from vnet_manager.conf import settings
from vnet_manager.config.config import get_config
from vnet_manager.config.validate import ValidateConfig
from vnet_manager.utils.version import show_version
from vnet_manager.utils.user import request_confirmation, generate_bash_completion_script
from vnet_manager.utils.files import write_file_to_disk, get_yaml_files_from_disk_path
from vnet_manager.environment.lxc import ensure_vnet_lxc_environment, cleanup_vnet_lxc_environment
from vnet_manager.operations.image import destroy_lxc_image
from vnet_manager.operations.files import put_files_on_machine, generate_vnet_hosts_file, place_vnet_hosts_file_on_machines
from vnet_manager.actions.help import display_help_for_action
from vnet_manager.operations.machine import (
    show_status,
    change_machine_status,
    create_machines,
    destroy_machines,
    enable_type_specific_machine_configuration,
)
from vnet_manager.operations.interface import (
    bring_up_vnet_interfaces,
    bring_down_vnet_interfaces,
    delete_vnet_interfaces,
    show_vnet_interface_status,
    show_vnet_veth_interface_status,
)

logger = getLogger(__name__)


class ActionManager:
    """
    The VNet Action Manager class
    Use this class to initiate specific VNet action in a controlled manner
    """

    def __init__(self, config_path=None, sniffer=False, base_image=False, no_hosts=False):
        """
        :param str config_path: The path to the config
        :param bool sniffer: Whether to enable sniffers on 'start'
        :param bool base_image: Whether to delete the base image on 'destroy'
        """
        self.config_path = config_path
        self.config = None
        self.sniffer = sniffer
        self.base_image = base_image
        self.no_hosts = no_hosts
        self._machines = None
        self._config_validated = False

    @property
    def machines(self):
        return self._machines

    @machines.setter
    def machines(self, machines):
        """
        Only preform actions on a certain amount of machines.
        Note that this only applies to the machine actions, all general actions will ignore this value.
        :param tpl/list: machines: The machine names to preform actions on, defaults to all
        """
        self._machines = machines

    def execute(self, action):
        """
        Execute an action
        :param str action: The name of the action to execute
        :raises NotImplementedError: If the action is unknown
        :raises RuntimeError
        :return: int: returncode of the action
        """
        # First do a sanity check on the action
        if action not in settings.VALID_ACTIONS:
            raise NotImplementedError("{} is not a valid action".format(action))
        # Check if the operator only wants to see the help text
        if self.config_path and self.config_path.lower() == "help":
            display_help_for_action(action)
            return EX_OK
        # Check if a config is required
        if action in settings.CONFIG_REQUIRED_ACTIONS:
            if not self.config_path:
                raise RuntimeError("The {} action requires a config to be passed, but it wasn't".format(action))
            if not self.parse_config():
                logger.critical("Config NOT OK, can't proceed")
                return EX_USAGE
        # Preform the action
        logger.info("Initiating {} action".format(action))
        getattr(self, "preform_{}_action".format(action.replace("-", "_")))()
        return EX_OK

    def parse_config(self):
        """
        Parses the user config
        Updates the config accordingly
        :return: bool: True if config parsing successful, False otherwise
        """
        self.config = get_config(self.config_path)
        # Config passed and required, validate it
        check_result, self.config = self.check_and_update_config()
        if not check_result:
            # Config NOT OK, quit execution and return usage code
            return False
        return True

    def check_and_update_config(self):
        """
        Validates the config passed to this instance.
        The validator can also fix some minor config issues, these are returned.
        :return: bool: True if config successfully validated, False otherwise / dict: The updated config
        """
        validator = ValidateConfig(self.config)
        validator.validate()
        if not validator.config_validation_successful:
            logger.error("The config seems to have unrecoverable issues, please fix them before proceeding")
            return False, dict()
        # Everything okay
        logger.debug("Config validation successful")
        return True, validator.updated_config

    def preform_show_action(self):
        show_status(self.config)
        show_vnet_interface_status(self.config)
        if "veths" in self.config:
            show_vnet_veth_interface_status(self.config)

    def preform_start_action(self):
        bring_up_vnet_interfaces(self.config, sniffer=self.sniffer)
        change_machine_status(self.config, machines=self._machines, status="start")

    def preform_stop_action(self):
        change_machine_status(self.config, machines=self._machines, status="stop")
        # If specific machines are specified, we don't want to mess with the interfaces
        if self._machines:
            logger.warning("Not bringing down VNet interfaces as we are only stopping specific machines, this may leave lingering sniffers")
        else:
            bring_down_vnet_interfaces(self.config)

    def preform_create_action(self):
        # Make sure the provider environments are correct
        ensure_vnet_lxc_environment(self.config)
        # Make the machines
        create_machines(self.config, machines=self._machines)
        # Put user requested file on the machines
        put_files_on_machine(self.config)
        if not self.no_hosts:
            # Put /etc/hosts on the machines
            generate_vnet_hosts_file(self.config)
            place_vnet_hosts_file_on_machines(self.config)
        # Configure type specific stuff
        enable_type_specific_machine_configuration(self.config)

    def preform_destroy_action(self):
        if self.base_image:
            request_confirmation(prompt="Are you sure you want to delete the VNet base images (y/n)? ")
            destroy_lxc_image(settings.LXC_BASE_IMAGE_ALIAS, by_alias=True)
        else:
            destroy_machines(self.config, machines=self._machines)
            # If specific machines are specified, we don't want to mess with the interfaces
            if self._machines:
                logger.warning(
                    "Not deleting VNet interfaces as we are only destroying specific machines, this may leave lingering sniffers"
                )
            else:
                delete_vnet_interfaces(self.config)

    def preform_list_action(self):
        # First check if we have been passed a file or directory
        if isfile(self.config_path):
            dep_msg = "List action with a regular config file is deprecated, use the 'show' action instead"
            logger.warning(dep_msg)
            warn(dep_msg)
            # Execute the show action instead
            self.execute("show")
        elif isdir(self.config_path):
            # We exclude the default.yaml config file because it is not a valid user config
            yaml_files = get_yaml_files_from_disk_path(self.config_path)
            for path in yaml_files:
                self.config_path = path
                if not self.parse_config():
                    logger.error("Config {} does not seem to be a valid config, skipping".format(path))
                    continue
                logger.info("Showing machine status for {}".format(path))
                show_status(self.config)
        else:
            logger.error(
                "Path {} does not seem to be a file or a directory, did you forget to pass a config directory?".format(self.config_path)
            )

    @staticmethod
    def preform_version_action():
        show_version()

    @staticmethod
    def preform_bash_completion_action():
        bash_script_content = generate_bash_completion_script()
        write_file_to_disk(settings.VNET_BASH_COMPLETION_PATH, bash_script_content)
        logger.info("Bash completion generated and placed. Use ' . /etc/bash_completion' to load in this terminal")

    @staticmethod
    def preform_clean_action():
        cleanup_vnet_lxc_environment()



import sys
from pathlib import Path
from os.path import join
from os import getenv

PYTHON_PACKAGE_NAME = "vnet-manager"
PROJECT_DIR = Path(__file__).resolve().parent.parent.parent
YAMLLINT_CONFIG_FILE = join(PROJECT_DIR, "yamllint.yaml")
CONFIG_FILE_DIR = join(PROJECT_DIR, "config")

# Log settings
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "console": {
            "()": "vnet_manager.utils.logging.formatters.ConsoleFormatter",
            "fmt": "%(asctime)s [%(name)8s] [%(levelname)s] %(message)s",
            "colored": sys.stderr.isatty,  # StreamHandler uses stderr by default
        },
        "syslog": {"format": "[%(process)d] %(name)s [%(levelname)s]: %(message)s"},
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "console",},
        "syslog": {
            "class": "logging.handlers.SysLogHandler",
            "formatter": "syslog",
            "address": "/dev/log",  # remove this to use UDP port 514
        },
    },
    "root": {"handlers": ["console", "syslog"], "level": "DEBUG",},
    # Silence debug heavy loggers here
    "loggers": {
        "urllib3": {"level": "INFO",},
        "ws4py": {"level": "WARNING",},
        "pyroute2": {"level": "INFO"},
        "pyroute2.ndb": {"level": "WARNING"},
    },
}

# VNet Manager static settings / config
# provider config
# The provider config defines some parameters that are used by VNet-manager for configuration. Most notably, the provider settings.
# Below the provider, config options are defined.
PROVIDERS = {
    "lxc": {
        "supported_operating_systems": ["bionic", "focal"],  # Will not run on any other OSes
        "dns-nameserver": "8.8.8.8",
        "required_host_packages": ["lxd", "lxc", "bridge-utils", "tcpdump", "net-tools", "curl"],  # List of packages required on the host
        "guest_packages": [
            "man",  # List of packages to install on the guest
            "net-tools",
            "traceroute",
            "nano",
            "vim",
            "bridge-utils",
            "radvd",
            "frr",
            "frr-pythontools",
            "vlan",
        ],
        "base_image": {  # Download info for the base image
            "os": "18.04",
            "server": "https://cloud-images.ubuntu.com/daily",
            "protocol": "simplestreams",
        },
    }
}
# The 'list' action also requires a config, but because it is handled differently we don't add it to CONFIG_REQUIRED_ACTIONS
CONFIG_REQUIRED_ACTIONS = ["show", "start", "stop", "status", "create", "destroy"]
VALID_ACTIONS = CONFIG_REQUIRED_ACTIONS + ["list", "clean", "version", "bash-completion"]
HELP_TEXT_ACTION_MAPPING = {
    "list": """Lists the status of the config files in a particular directory.
Usage: vnet-manager list <dir>.
In which the <dir> contains vnet-manager user config yaml files.
This command does a recursive search for all vnet-manager config in the supplied directory (so it also lists subdirectories).
No interface status will be shown, as these might overlap for some configs.
    """,
    "show": """Show the current status of the supplied config file
    """,
    "start": """Starts up a previously built config.
This action will also bring up/start required VNet interfaces and enable sniffers if the --sniffer option is passed.
    """,
    "stop": """Stops a previously started config.
This action will also bring down the corresponding VNet interfaces.
    """,
    "status": """Shows the current status of the supplied config file.
    """,
    "destroy": """Destroys a previously built config.
This action will delete the corresponding machines and VNet interfaces.
Use this action in combination with the --base-image parameter to delete the VNet base images.
    """,
    "clean": """Purge VNet specific provider configuration from the system.
This action can only if all VNet config on the system are destroyed.
Only use this action when you wish to reset or uninstall VNet manager.
    """,
    "create": """Builds a VNet configuration so it can be started using the 'start' action.
This action will create a base image for all providers present in the configuration if they do not exist yet.
    """,
    "bash-completion": """Places the VNet-manager bash completion script""",
}
VNET_BRIDGE_NAME = "vnet-br"
VNET_SNIFFER_PCAP_DIR = getenv("VNET_SNIFFER_PCAP_DIR", "/tmp")
SUPPORTED_MACHINE_TYPES = ["host", "router"]
MACHINE_TYPE_PROVIDER_MAPPING = {
    "host": "lxc",
    "router": "lxc",
}
MACHINE_TYPE_CONFIG_FUNCTION_MAPPING = {
    # For each machine type specify the type specific functions that should be called for it
    "host": ["disable_{}_ip_forwarding".format(MACHINE_TYPE_PROVIDER_MAPPING["router"])],
    "router": ["enable_{}_ip_forwarding".format(MACHINE_TYPE_PROVIDER_MAPPING["router"])],
}
VALID_STATUSES = ["start", "stop"]
VNET_FORCE_ENV_VAR = "VNET_FORCE"
VNET_ETC_HOSTS_FILE_PATH = "/tmp/.vnet_etc_hosts"
VNET_STATIC_HOSTS_FILE_PART = """
127.0.0.1 localhost

# The following lines are desirable for IPv6 capable hosts
::1 ip6-localhost ip6-loopback
fe00::0 ip6-localnet
ff00::0 ip6-mcastprefix
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
ff02::3 ip6-allhosts

# VNet hosts

"""
VNET_NETPLAN_CONFIG_FILE_PATH = "/etc/netplan/10-vnet-config.yaml"
VNET_BASH_COMPLETION_TEMPLATE = """#!/usr/bin/env bash

_{name}_completions() {{
    local opts
    opts="{options}"
    case $COMP_CWORD in
        1)
            COMPREPLY=( $(compgen -W "${{opts}}" -- "${{COMP_WORDS[COMP_CWORD]}}") )
            ;;
        2)
            COMPREPLY=( $(compgen -o default -- "${{COMP_WORDS[COMP_CWORD]}}") )
            ;;
    esac
    return 0
}}
complete -F _{name}_completions {name}
"""
VNET_BASH_COMPLETION_PATH = "/etc/bash_completion.d/{}.bash".format(PYTHON_PACKAGE_NAME)

# LXC specific settings
LXC_MAX_STATUS_WAIT_ATTEMPTS = 15
LXC_STATUS_WAIT_SLEEP = 4
LXC_STATUS_BACKOFF_MULTIPLIER = 0.3
LXC_STORAGE_POOL_NAME = "vnet-pool"
LXC_STORAGE_POOL_DRIVER = "btrfs"
LXC_STORAGE_POOL_SIZE = "30GB"
LXC_BASE_IMAGE_ALIAS = getenv("VNET_LXC_BASE_IMAGE", "vnet-base-image")
LXC_BASE_IMAGE_MACHINE_NAME = "vnet-base"
LXC_VNET_PROFILE = "vnet-profile"

# FRR settings
FRR_RELEASE = "frr-stable"



from unittest import TestCase, mock
from pathlib import Path

from vnet_manager.conf import settings


class VNetTestCase(TestCase):
    def __init__(self, *args, **kwargs):
        self.expected_git_path = Path(settings.PROJECT_DIR, ".git")
        self.assertTrue(self.expected_git_path.is_dir(), "project_path is not correct? (no .git found)")
        super(VNetTestCase, self).__init__(*args, **kwargs)

    def set_up_patch(self, topatch, themock=None, **kwargs):
        """
        Patch a function or class
        :param topatch: string The class to patch
        :param themock: optional object to use as mock
        :return: mocked object
        """
        if themock is None:
            themock = mock.Mock()

        if "return_value" in kwargs:
            themock.return_value = kwargs["return_value"]

        patcher = mock.patch(topatch, themock)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def set_up_context_manager_patch(self, topatch, themock=None, **kwargs):
        """
        Provides a mock object which can be used with context managers (like with statements)
        """
        patcher = self.set_up_patch(topatch, themock=themock, **kwargs)
        patcher.return_value.__exit__ = lambda a, b, c, d: None
        patcher.return_value.__enter__ = patcher
        return patcher



from logging import makeLogRecord, Formatter, WARNING

from vnet_manager.tests import VNetTestCase
from vnet_manager.utils.logging.formatters import ConsoleFormatter


class TestConsoleFormatter(VNetTestCase):
    def setUp(self) -> None:
        self.record = makeLogRecord({"name": "salty.sub", "levelno": WARNING, "levelname": "WARNING", "msg": "mocked log",})
        self.formatter = ConsoleFormatter()
        self.fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    def test_console_formatter_extends_formatter(self):
        self.assertIsInstance(self.formatter, Formatter)

    def test_console_formatter_formats_when_not_colored_with_default_fmt(self):
        formatter = ConsoleFormatter(colored=False)
        self.assertEqual(formatter.format(self.record), Formatter().format(self.record))

    def test_console_formatter_formats_when_not_colored_with_custom_fmt(self):
        formatter = ConsoleFormatter(self.fmt, colored=False)
        self.assertEqual(formatter.format(self.record), Formatter(self.fmt).format(self.record))

    def test_console_formatter_formats_when_colored_with_default_fmt(self):
        colorless_msg = Formatter().format(self.record)
        colored_msg = ConsoleFormatter(colored=True).format(self.record)

        self.assertNotEqual(colorless_msg, colored_msg)
        for word in colorless_msg.split():
            self.assertIn(word, colored_msg)

    def test_console_formatter_formats_when_colored_with_custom_fmt(self):
        colorless_msg = Formatter(self.fmt).format(self.record)
        colored_msg = ConsoleFormatter(self.fmt, colored=True).format(self.record)

        self.assertNotEqual(colorless_msg, colored_msg)
        for word in colorless_msg.split():
            self.assertIn(word, colored_msg)

    def test_console_formatter_accepts_functions_to_detect_colored(self):
        is_colored_calls = []

        def is_colored():
            is_colored_calls.append(True)
            return True

        colorless_msg = Formatter().format(self.record)
        colored_msg = ConsoleFormatter(colored=is_colored).format(self.record)

        self.assertNotEqual(colorless_msg, colored_msg)
        for word in colorless_msg.split():
            self.assertIn(word, colored_msg)

        self.assertEqual(is_colored_calls, [True])

    def test_console_formatter_wont_color_if_colored_func_returns_false(self):
        formatter = ConsoleFormatter(colored=lambda: False)
        self.assertEqual(formatter.format(self.record), Formatter().format(self.record))

    def test_console_formatter_does_not_accept_args_other_than_bool_or_callable(self):
        with self.assertRaises(TypeError):
            ConsoleFormatter(colored="i am not bool nor function")
